'
'	An example of compressing or decompressing files using zlib.
' For info on zlib see https://zlib.net/
'
'	Michael McElligott
'	12'th of April 20004
'
'
PROGRAM	"xpack"
VERSION	"0.10"
CONSOLE


	IMPORT "kernel32"
	IMPORT "msvcrt"
	IMPORT "zlib"	
	IMPORT "xio"

$$xpk_ident		= "xpk"
$$xpk_version	= 1
$$xpk_none		= 1
$$xpk_zlib		= 2

PACKED TXPAK
	STRING * 3	.xpakid		' xpack file head identifier
	UBYTE		.version			' xpack version
	UBYTE		.comptype			' compression type: 1 = none, 2 = zlib
	'UBYTE	.encrypt			' TODO
	USHORT	.fnlen				' length of filename which follows this strut + \n
	ULONG		.totalsize		' size of data before compression
	ULONG		.compsize			' size of data after compression
	ULONG		.adler				' adler checksum of data before compression
END TYPE

DECLARE FUNCTION Entry ()
DECLARE FUNCTION zlib_decompress (source,ssize,dest,dsize)
DECLARE FUNCTION zlib_compress (source,ssize,dest,dsize,level)
DECLARE FUNCTION open_file (lpfilename, flags)
DECLARE FUNCTION close_file (hfile)
DECLARE FUNCTION write_file (hfile,ULONG buffer,nbytes)
DECLARE FUNCTION read_file (hfile,ULONG buffer,nbytes)
DECLARE FUNCTION flof (hfile)
DECLARE FUNCTION create_buffer (size) 
DECLARE FUNCTION free_buffer (buffer) 
DECLARE FUNCTION GetAdler32 (buffer,tbytes)
DECLARE FUNCTION getLastSlash(str$, stop)
DECLARE FUNCTION DecomposePathname (pathname$, @path$, @parent$, @filename$, @file$, @extent$)

EXPORT
'
DECLARE FUNCTION compress_file (STRING srcfile, STRING destfile, UBYTE comptype, UBYTE level)
' if destfile is empty then srcfile will be used with the extension ".xpk"

DECLARE FUNCTION decompress_file (STRING srcfile, STRING destfile)
' if destfile is empty then original filename will be used
END EXPORT


FUNCTION Entry ()

	IF LIBRARY (0) THEN RETURN

	compress_file ("xpack.x", "", $$xpk_zlib, $$Z_BEST_COMPRESSION)
	PRINT "compressed xpack.x to xpack.xpk"
	
	decompress_file ("xpack.xpk", "_xpack.x")
	PRINT "decompressed xpack.xpk to _xpack.x"
	
	INLINE$ ("Press any key to quit >")

END FUNCTION

FUNCTION decompress_file (STRING srcfile,STRING destfile)
	TXPAK xpak
	STRING origfile
	
	
	hfile = open_file (&srcfile,0)
	IFZ hfile THEN RETURN 0
	
	read_file (hfile,&xpak,SIZE(TXPAK))
	IF (xpak.xpakid != $$xpk_ident) THEN
		printf(&"'%s' is not a xpk file\n",&srcfile)
		RETURN 0
	END IF
	
	origfile = NULL$(xpak.fnlen)
	sbuffer = create_buffer (xpak.compsize)
	read_file (hfile,&origfile,xpak.fnlen)
	read_file (hfile,sbuffer,xpak.compsize)
	close_file (hfile)

	origfile = TRIM$(origfile)
	IFZ destfile THEN destfile = origfile 	

	hfile = open_file (&destfile,&"wb")
	IFZ hfile THEN RETURN 0
	SELECT CASE xpak.comptype
		CASE $$xpk_zlib		:
			dbuffer = create_buffer (xpak.totalsize + 50)
			zlib_decompress (sbuffer,xpak.compsize,dbuffer,&xpak.totalsize)
			IF (GetAdler32(dbuffer,xpak.totalsize) != xpak.adler) THEN
				printf (&"Adler checksum mismatch on '%s'\n",&origfile)
				'continue anyway
			END IF
			write_file (hfile,dbuffer,xpak.totalsize)
			free_buffer (dbuffer)

		CASE $$xpk_none		:
			IF (GetAdler32(sbuffer,xpak.totalsize) != xpak.adler) THEN
				printf (&"Adler checksum mismatch on '%s'\n",&origfile)
			END IF
			write_file (hfile,sbuffer,xpak.totalsize)
			free_buffer (sbuffer)
	END SELECT
	close_file (hfile)

END FUNCTION 

FUNCTION compress_file (STRING srcfile,STRING destfile,UBYTE comptype,UBYTE level)
	TXPAK xpak

	xpak.xpakid = $$xpk_ident
	xpak.version = $$xpk_version
	xpak.comptype = comptype

	hfile = open_file (&srcfile,0)
	IFZ hfile THEN RETURN 0
	xpak.totalsize = flof (hfile)
	
	sbuffer = create_buffer (xpak.totalsize)
	xpak.compsize = (read_file(hfile,sbuffer,xpak.totalsize)) + 50
	xpak.adler = GetAdler32 (sbuffer,xpak.totalsize)
	close_file (hfile)

	srcfile = srcfile + "\n"
	xpak.fnlen = SIZE(srcfile)
	
	IFZ destfile THEN
		DecomposePathname (srcfile, path$, parent$, filename$, @file$, extent$)
		destfile = file$ +".xpk"
	END IF
	
	hfile = open_file (&destfile,&"wb")
	IFZ hfile THEN RETURN 0
	write_file (hfile,&xpak,SIZE(TXPAK))
	write_file (hfile,&srcfile,xpak.fnlen)	
	
	SELECT CASE comptype
		CASE $$xpk_zlib		:
			dbuffer = create_buffer (xpak.compsize)	
			zlib_compress (sbuffer,xpak.totalsize,dbuffer,&xpak.compsize,level)
			free_buffer (sbuffer)
			write_file (hfile,dbuffer,xpak.compsize)
			free_buffer (dbuffer)
			
		CASE $$xpk_none		:
			write_file (hfile,sbuffer,xpak.totalsize)
			free_buffer (sbuffer)
	END SELECT
	
	close_file (hfile)

END FUNCTION

FUNCTION GetAdler32 (buffer,tbytes)

	adler = adler32(0, $$Z_NULL, 0)
	RETURN adler32(adler,buffer,tbytes)
END FUNCTION

FUNCTION create_buffer (nbytes)

	buffer = malloc (nbytes)
	IFZ buffer THEN RETURN 0
	RtlZeroMemory (buffer,nbytes)
	
'	buffer = GlobalAlloc ($$GHND, nbytes)
'	IFZ hAlloc THEN RETURN 
'	buffer = GlobalLock (hAlloc)


	RETURN buffer
END FUNCTION

FUNCTION free_buffer (buffer) 

	GlobalFree (buffer)
'	RETURN free (buffer)
END FUNCTION

FUNCTION flof (hfile)

	pos = 0
	fgetpos (hfile,&pos)
	fseek(hfile, 0, 2)
	size = ftell(hfile)
	fsetpos (hfile,&pos)
	
	RETURN size
END FUNCTION

FUNCTION zlib_decompress (source,ssize,dest,dsize)

	IF (uncompress (dest,dsize,source,ssize) == $$Z_OK) THEN
		RETURN $$TRUE
	ELSE
		printf(&"zlib_decompress(): unable to decompress buffer\n",$$NULL)
		RETURN $$FALSE
	END IF
END FUNCTION

FUNCTION zlib_compress (source,ssize,dest,dsize,level)

	IF (compress2 (dest,dsize,source,ssize,level) == $$Z_OK) THEN
		RETURN $$TRUE
	ELSE
		printf(&"zlib_compress(): unable to compress buffer\n",$$NULL)
		RETURN $$FALSE
	END IF
END FUNCTION

FUNCTION open_file (lpfilename, flags)

	IFZ lpfilename THEN RETURN $$FALSE
	IFZ flags THEN
		type = &"rb"
	ELSE
		type = flags
	END IF
	
	hfile = fopen (lpfilename, type)
	IFZ (hfile)
		printf(&"unable to open file '%s'\n",lpfilename)
		RETURN 0
	ELSE
		RETURN hfile
	END IF

END FUNCTION

FUNCTION close_file (hfile)

	IF hfile THEN
		fclose (hfile)
		RETURN $$TRUE
	ELSE
		printf(&"unable to close file\n",NULL)
		RETURN $$FALSE
	END IF

END FUNCTION

FUNCTION write_file (hfile,ULONG buffer,nbytes)

	'_write (hfile, buffer, nbytes)
	IF (fwrite (buffer, 1, nbytes, hfile) < nbytes) THEN
		printf(&"error writing buffer to file\n",NULL)
		RETURN $$FALSE
	ELSE
		RETURN $$TRUE
	END IF
END FUNCTION

FUNCTION read_file (hfile,ULONG buffer,nbytes)

	tbytes = fread  (buffer, 1, nbytes, hfile)
	IF (tbytes < nbytes) THEN
		IF feof(hfile) THEN
			RETURN tbytes
		ELSE
			printf(&"error reading file to buffer\n",NULL)
			RETURN 0
		END IF
	ELSE
		RETURN tbytes
	END IF
END FUNCTION


FUNCTION DecomposePathname (pathname$, @path$, @parent$, @filename$, @file$, @extent$)
'
	path$ = ""
	file$ = ""
	extent$ = ""
	parent$ = ""
	filename$ = ""
	name$ = TRIM$ (pathname$)
	dot = RINSTR (name$, ".")
	slash = getLastSlash(name$, -1)
	
	IF slash THEN preslash = getLastSlash(name$, slash-1)
	IF (dot < slash) THEN dot = 0
'
	filename$ = MID$ (name$, slash+1)							' filename = "name.ext"
	IFZ dot THEN
		file$ = filename$										' file = filename (filename has no extent)
	ELSE
		file$ = MID$ (name$, slash+1, dot-slash-1)	' file = "name" (without extent)
		extent$ = MID$ (name$, dot)					' extent = ".ext"
	END IF
'
	IF slash THEN
		path$ = LEFT$ (name$, slash-1)							' path = full pathname to left of "/file.ext"
		IF preslash THEN
			parent$ = MID$ (name$, preslash+1, slash-preslash-1)
		ELSE
			parent$ = LEFT$ (name$, slash-1)
		END IF
	END IF
	
END FUNCTION

FUNCTION getLastSlash(str$, stop)
	$PathSlash$   = "\\" 

	IF stop < 0 THEN
		slash1 = RINSTR(str$, "/")
		slash2 = RINSTR(str$, $PathSlash$)
	ELSE
		slash1 = RINSTR(str$, "/", stop)
		slash2 = RINSTR(str$, $PathSlash$, stop)
	END IF
	IFZ slash1 THEN
		RETURN slash2
	ELSE
		RETURN MAX(slash1, slash2)
	END IF
	
END FUNCTION

END PROGRAM

